using Jypeli;
using Jypeli.Assets;
using Jypeli.Controls;
using Jypeli.Widgets;
using System;
using System.Collections.Generic;

/// @author Roy Eskelinen
/// @version 6.1.2022
/// <summary>
/// Eskape tasohyppelypeli.
/// Silmukkatehtävä: https://tim.jyu.fi/view/kurssit/tie/ohj1/2021s/demot/demo5#teht%C3%A4v%C3%A4-4.-rajan-ylitt%C3%A4v%C3%A4t
/// Taulukkotehtävä: https://tim.jyu.fi/view/kurssit/tie/ohj1/2021s/demot/demo10#v2 
/// </summary>


public class Eskape : PhysicsGame
{
    private const double NOPEUS = 150; // hahmon nopeus
    private const double HYPPYNOPEUS = 900; // kuinka korkealle hahmo hyppää
    private const int RUUDUN_KOKO = 40; // ruudun zoomaus

    private PlatformCharacter pelaaja;

    private int healthPoints = 5; // hahmon elämät
    private IntMeter pistelaskuri = new IntMeter(0);

    // kuvat
    private Image pelaajanKuva = LoadImage("PlayerHero1.png");
    private Image enemyPic = LoadImage("Enemy1.png");
    private Image enemyPic2 = LoadImage("Enemy2.png");
    private Image point1 = LoadImage("points.png");
    private Image point2 = LoadImage("points 2.png");
    private Image boxPic = LoadImage("Box3.png");
    private Image floorPic = LoadImage("floor.png");
    private Image invisibleFloorPic = LoadImage("nakymatonTaso.png");
    private Image spikesPic = LoadImage("spikes.png");
    private Image healthPotionPic = LoadImage("healthpotion.png");
    private Image taustakuva = LoadImage("background2.jpg");
    private Image bullet = LoadImage("bullet.png");
    private Image skull = LoadImage("skull.png");

    // äänet
    private SoundEffect hyppyAani = LoadSoundEffect("Jump3.wav");
    private SoundEffect ampuminenAani = LoadSoundEffect("Laser_Shoot.wav");
    private SoundEffect rajahdysAani = LoadSoundEffect("explosion.wav");
    private SoundEffect osumaAani = LoadSoundEffect("Hit_Hurt3.wav");
    private SoundEffect coinAani = LoadSoundEffect("pickCoin.wav");
    private SoundEffect hpAani = LoadSoundEffect("powerUp.wav");


    /// <summary>
    /// Begin kutsuu aliohjelman AloitaPelin.
    /// </summary>
    public override void Begin()
    {
        AloitaPeli();
    }


    /// <summary>
    /// Kentän sisältö. Perässä olevat nimet viittaavat kutsuttaviin aliohjelmiin.
    /// </summary>
    private void LuoKentta()
    {
        TileMap kentta = TileMap.FromLevelAsset("kentta1.txt");
        kentta.SetTileMethod('#', LisaaTaso);
        kentta.SetTileMethod('X', LisaaNakymatonTaso);
        kentta.SetTileMethod('T', LisaaTappavaTaso);
        kentta.SetTileMethod('*', LisaaPiste);
        kentta.SetTileMethod('^', LisaaPiste2);
        kentta.SetTileMethod('N', LisaaPelaaja);
        kentta.SetTileMethod('E', LisaaVihollinen);
        kentta.SetTileMethod('F', LisaaVihollinen);
        kentta.SetTileMethod('B', LisaaLaatikko);
        kentta.SetTileMethod('S', LisaaPommi);
        kentta.SetTileMethod('H', LisaaHealthPotion);
        kentta.Execute(RUUDUN_KOKO, RUUDUN_KOKO);
        Level.CreateBorders();
        Level.Background.Image = taustakuva;
        Level.Background.TileToLevel();
    }


    /// <summary>
    /// Health Potion. Sen keräämällä hahmo saa yhden elämäpisteen lisää.
    /// </summary>
    /// <param name="paikka">Potionin paikka. Määritellään kentässä.</param>
    /// <param name="leveys">Koko leveyssuunnassa.</param>
    /// <param name="korkeus">Koko pituussuunnassa.</param>
    void LisaaHealthPotion(Vector paikka, double leveys, double korkeus)
    {
        PhysicsObject healthPotion = PhysicsObject.CreateStaticObject(leveys, korkeus);
        healthPotion.IgnoresCollisionResponse = true;
        healthPotion.Image = healthPotionPic;
        healthPotion.Position = paikka;
        healthPotion.Tag = "health";
        Add(healthPotion);
    }


    /// <summary>
    /// Pelin perustaso, jonka päällä kävellään.
    /// </summary>
    /// <param name="paikka">Tason paikka. Määritellään kentta1.txt:ssä</param>
    /// <param name="leveys">Tason leveys.</param>
    /// <param name="korkeus">Tason korkeus</param>
    private void LisaaTaso(Vector paikka, double leveys, double korkeus)
    {
        PhysicsObject taso = PhysicsObject.CreateStaticObject(leveys + 1, korkeus + 1);
        taso.Position = paikka;
        taso.Image = floorPic;
        taso.Tag = "floor";
        Add(taso);
    }


    /// <summary>
    /// Näkymätön taso pelissä. Käyttäytyy samalla tavalla kuin "LisaaTaso"-aliohjelma, mutta on näkymätön.
    /// </summary>
    /// <param name="paikka">Sijainti kentässä.</param>
    /// <param name="leveys">Leveys</param>
    /// <param name="korkeus">Korkeus</param>
    private void LisaaNakymatonTaso(Vector paikka, double leveys, double korkeus)
    {
        PhysicsObject taso = PhysicsObject.CreateStaticObject(leveys, korkeus);
        taso.Position = paikka;
        taso.Image = invisibleFloorPic;
        taso.Tag = "floor";
        Add(taso);
    }


    /// <summary>
    /// Pelihahmon tappava taso. Tason päälle tippuessaan peli alkaa välittömästi alusta.
    /// </summary>
    /// <param name="paikka">Tappavan tason paikka kentässä.</param>
    /// <param name="leveys">Tappavan tason leveys.</param>
    /// <param name="korkeus">Tappavan tason korkeus.</param>
    private void LisaaTappavaTaso(Vector paikka, double leveys, double korkeus)
    {
        PhysicsObject spikes = PhysicsObject.CreateStaticObject(leveys, korkeus);
        spikes.Position = paikka;
        spikes.Image = spikesPic;
        spikes.Tag = "spikes";
        Add(spikes);
    }


    /// <summary>
    /// Pelissä kerättävä kolikko. Antaa yhden pisteen.
    /// </summary>
    /// <param name="paikka">Kolikon paikka kentässä.</param>
    /// <param name="leveys">Kolikon leveys.</param>
    /// <param name="korkeus">Kolikon korkeus.</param>
    private void LisaaPiste(Vector paikka, double leveys, double korkeus)
    {
        PhysicsObject piste = PhysicsObject.CreateStaticObject(leveys, korkeus);
        piste.IgnoresCollisionResponse = true;
        piste.Position = paikka;
        piste.Image = point1;
        piste.Tag = "piste";
        Add(piste);
    }


    /// <summary>
    /// Pelissä kerättävä kolikko. Antaa viisi pistettä.
    /// </summary>
    /// <param name="paikka">Kolikon paikka kentässä.</param>
    /// <param name="leveys">Kolikon leveys.</param>
    /// <param name="korkeus">Kolikon korkeus.</param>
    private void LisaaPiste2(Vector paikka, double leveys, double korkeus)
    {
        PhysicsObject piste2 = PhysicsObject.CreateStaticObject(leveys, korkeus);
        piste2.IgnoresCollisionResponse = true;
        piste2.Position = paikka;
        piste2.Image = point2;
        piste2.Tag = "piste2";
        Add(piste2);
    }


    /// <summary>
    /// Pelihahmo, jota ohjataan.
    /// </summary>
    /// <param name="paikka">Aloituspaikka pelissä.</param>
    /// <param name="leveys">Hahmon leveys.</param>
    /// <param name="korkeus">Hahmon korkeus.</param>
    public void LisaaPelaaja(Vector paikka, double leveys, double korkeus)
    {
        pelaaja = new PlatformCharacter(35, 39);
        pelaaja.Position = paikka;
        pelaaja.Image = pelaajanKuva;
        pelaaja.Weapon = new AssaultRifle(0, 0);
        pelaaja.Weapon.Ammo.Value = 999;
        pelaaja.Weapon.ProjectileCollision = AmmusOsui;
        pelaaja.Tag = "pelaaja";

        // Alla aliohjelmat pelaajan eri reaktioihin kentän kanssa.
        AddCollisionHandler(pelaaja, "piste", TormaaKolikkoon);
        AddCollisionHandler(pelaaja, "piste2", TormaaKolikkoon2);
        AddCollisionHandler(pelaaja, "tappavat", TormaaPommiin);
        AddCollisionHandler(pelaaja, "health", TormaaHealthPotion);
        AddCollisionHandler(pelaaja, "spikes", TormaaTappavaanTasoon);

        Add(pelaaja);
    }


    /// <summary>
    /// Pelissä summittaisesti lentelevä pommi, joka osuessaan pelaajaan satuttaa pelaajaa 1 pisteen verran.
    /// </summary>
    /// <param name="paikka">Aloituspaikka pommille. Muuttuu välittömästi pelin alkaessa.</param>
    /// <param name="leveys">Pommin leveys.</param>
    /// <param name="korkeus">Pommin korkeus.</param>
    private void LisaaPommi(Vector paikka, double leveys, double korkeus)
    {
        PhysicsObject paakallo = new PhysicsObject(30, 30);

        RandomMoverBrain satunnaisaivot = new RandomMoverBrain(150); // tekoäly
        satunnaisaivot.ChangeMovementSeconds = 5;
        paakallo.Brain = satunnaisaivot;
        satunnaisaivot.WanderRadius = 400;
        paakallo.Position = paikka;

        paakallo.Image = skull;
        paakallo.Tag = "tappavat";
        Add(paakallo);
    }


    /// <summary>
    /// Pelihahmoon lisättävä vihollinen, joka ampuu pelaajaa. Tuhotessaan pelaaja saa vihollisesta 10 pistettä.
    /// </summary>
    /// <param name="paikka">Vihollisen paikka.</param>
    /// <param name="leveys">Vihollisen leveys.</param>
    /// <param name="korkeus">Vihollisen korkeus.</param>
    private void LisaaVihollinen(Vector paikka, double leveys, double korkeus)
    {
        PlatformCharacter enemy1 = new PlatformCharacter(40, 40);

        enemy1.Mass = 8.0;
        enemy1.Position = paikka;
        enemy1.Weapon = new AssaultRifle(0, 0);
        enemy1.Weapon.Ammo.Value = 9999;
        enemy1.Image = RandomGen.SelectOne(enemyPic, enemyPic2);
        enemy1.Tag = "vihu";
        enemy1.CanRotate = false;
        Add(enemy1);

        // Vihollisen tekoäly.
        PlatformWandererBrain tasoAivot = new PlatformWandererBrain();
        tasoAivot.Speed = 100;
        tasoAivot.JumpSpeed = 200;
        tasoAivot.TriesToJump = false;
        enemy1.Brain = tasoAivot;

        // Alla oleva koodi määrittää vihollisen ampumisen.
        Timer.CreateAndStart(RandomGen.NextDouble(0.75, 1.75), delegate 
        {
            Vector d = pelaaja.Position - enemy1.Position;
            d.Y = 0;
            if (d.Magnitude > 200)
                return;
            enemy1.Weapon.Angle = d.Angle;
            AmmuAseella(enemy1, "vihuluoti");
        });
    }


    /// <summary>
    /// Pistelaskuri. Pistelaskuri reagoi pelihahmon interaktioon kentän kanssa.
    /// </summary>
    private void LuoPistelaskuri()
    {
        pistelaskuri = new IntMeter(0);

        Label pistenaytto = new Label();

        pistenaytto.X = Screen.Left + 100;
        pistenaytto.Y = Screen.Top - 100;
        pistenaytto.TextColor = Color.Yellow;
        pistenaytto.Color = Color.Black;
        pistenaytto.BindTo(pistelaskuri);
        Add(pistenaytto);
    }


    /// <summary>
    /// Peliin lisättävä objekti, joka ei tuhoudu, mutta jota pelihahmolla voidaan liikutella.
    /// </summary>
    /// <param name="paikka">Laatikon sijainti.</param>
    /// <param name="leveys">Laatikon leveys.</param>
    /// <param name="korkeus">Laatikon paikka.</param>
    private void LisaaLaatikko(Vector paikka, double leveys, double korkeus)
    {
        PhysicsObject box = new PhysicsObject(leveys, korkeus);
        box.Mass = 5.0;
        box.Position = paikka;
        box.Image = boxPic;
        Add(box);
    }


    /// <summary>
    /// Pelissä olevat näppäimet.
    /// </summary>
    private void LisaaNappaimet()
    {
        Keyboard.Listen(Key.F1, ButtonState.Pressed, ShowControlHelp, "Näytä ohjeet");
        Keyboard.Listen(Key.Escape, ButtonState.Pressed, ConfirmExit, "Lopeta peli");
        Keyboard.Listen(Key.Left, ButtonState.Down, Liikuta, "Liikkuu vasemmalle", pelaaja, -NOPEUS);
        Keyboard.Listen(Key.Right, ButtonState.Down, Liikuta, "Liikkuu vasemmalle", pelaaja, NOPEUS);
        Keyboard.Listen(Key.Up, ButtonState.Pressed, Hyppaa, "Pelaaja hyppää", pelaaja, HYPPYNOPEUS);
        Keyboard.Listen(Key.RightControl, ButtonState.Down, AmmuAseella, "Ammu", pelaaja, "pelaajaluoti");
        PhoneBackButton.Listen(ConfirmExit, "Lopeta peli");
    }


    /// <summary>
    /// Aseella ampumisen aliohjelma.
    /// </summary>
    /// <param name="pelaaja">Pelihahmo.</param>
    /// <param name="tunnus">Ase.</param>
    private void AmmuAseella(PlatformCharacter pelaaja, string tunnus)
    {
        PhysicsObject ammus = pelaaja.Weapon.Shoot();

        pelaaja.Weapon.FireRate = 2;
        pelaaja.Weapon.ProjectileCollision = AmmusOsui;

        ampuminenAani.Play();

        if (ammus != null)
        {
            ammus.Tag = tunnus;
            ammus.Size *= 3;
            ammus.Image = bullet;
            ammus.MaximumLifetime = TimeSpan.FromSeconds(7.0);
        }
    }


    /// <summary>
    /// Aliohjelma määrittää, mitä tapahtuu kun aseella osutaan.
    /// </summary>
    /// <param name="ammus">Luoti</param>
    /// <param name="kohde">Kohde, johon luoti osuu.</param>
    private void AmmusOsui(PhysicsObject ammus, PhysicsObject kohde)
    {
        if (kohde.Tag.ToString() == "vihu")
        {
            kohde.Destroy();
            pistelaskuri.AddValue(10);
        }

        if (kohde.Tag.ToString() == "pelaaja")
        {
            // MessageDisplay.Add("Osui"); Tällä testattiin, osuiko peli.
            healthPoints--;
        }
        osumaAani.Play();
        ammus.Destroy();
    }


    /// <summary>
    /// Pelihahmon liikuttaminen.
    /// </summary>
    /// <param name="hahmo">Pelihahmo.</param>
    /// <param name="nopeus">Pelihahmon nopeus.</param>
    private void Liikuta(PlatformCharacter hahmo, double nopeus)
    {
        if (hahmo == null || hahmo.IsDestroyed)
            return;
        hahmo.Walk(nopeus);
    }


    /// <summary>
    /// Pelihahmon hyppääminen.
    /// </summary>
    /// <param name="hahmo">Pelihahmo</param>
    /// <param name="nopeus">Hyppynopeus.</param>
    private void Hyppaa(PlatformCharacter hahmo, double nopeus)
    {
        hyppyAani.Play();
        hahmo.Jump(nopeus);
    }


    /// <summary>
    /// Reaktio silloin, kun pelihahmo törmää health potioniin. Pelihahmo saa tällöin +1 elämäpistettä lisää.
    /// </summary>
    /// <param name="hahmo">Pelihahmo</param>
    /// <param name="healthPotion">Health potion</param>
    private void TormaaHealthPotion(PhysicsObject hahmo, PhysicsObject healthPotion)
    {
        healthPoints++;
        hpAani.Play();
        MessageDisplay.Add("HP +1");
        healthPotion.Destroy();
    }


    /// <summary>
    /// Reaktio, kun pelihahmo osuu 1 pisteen kolikkoon.
    /// </summary>
    /// <param name="hahmo">Pelihahmo.</param>
    /// <param name="kolikko">Kolikko</param>
    private void TormaaKolikkoon(PhysicsObject hahmo, PhysicsObject kolikko)
    {
        pistelaskuri.AddValue(1);
        coinAani.Play();
        kolikko.Destroy();
        // pistelaskuri.Value += 1;
    }


    /// <summary>
    /// Reaktio, kun pelihahmo osuu 5 pisteen kolikkoon.
    /// </summary>
    /// <param name="hahmo">Pelihahmo.</param>
    /// <param name="kolikko">Kolikko</param>
    private void TormaaKolikkoon2(PhysicsObject hahmo, PhysicsObject kolikko)
    {
        coinAani.Play();
        kolikko.Destroy();
        pistelaskuri.AddValue(5);
    }


    /// <summary>
    /// Reaktio, kun pelihahmo osuu pommiin.
    /// </summary>
    /// <param name="hahmo">Pelihahmo</param>
    /// <param name="skull">Pommi</param>
    private void TormaaPommiin(PhysicsObject hahmo, PhysicsObject skull)
    {
        healthPoints--;

        if (healthPoints <= 0)
        {
            pelaaja.Destroy();
            Timer.SingleShot(3, AloitaPeli);
        }
        rajahdysAani.Play();
        MessageDisplay.Add("HP -1");
        skull.Destroy();
    }


    /// <summary>
    /// Reaktio, kun pelihahmo osuu tappavaan tasoon.
    /// </summary>
    /// <param name="hahmo">Pelihahmo.</param>
    /// <param name="spikes">Tappava taso.</param>
    private void TormaaTappavaanTasoon(PhysicsObject hahmo, PhysicsObject spikes)
    {
        healthPoints = 0;
        pelaaja.Destroy();
        Timer.SingleShot(3, AloitaPeli);
        MessageDisplay.Add("Kuolit");
        // lisää tuskanhuuto
    }


    /// <summary>
    /// Aliohjelma, joka aloittaa pelin alusta silloin, kun pelihahmo kuolee.
    /// </summary>
    private void AloitaPeli()
    {
        ClearAll();

        Gravity = new Vector(300, -2000);

        LuoKentta();
        LisaaNappaimet();
        LuoPistelaskuri();

        Camera.Follow(pelaaja);
        Camera.ZoomFactor = 2;
        Camera.StayInLevel = true;

        MasterVolume = 0.1;
    }
}

